#include <cmath>
#include <cstdlib>
#include <iostream>
#include <sstream>
#include <string>

using namespace std;
#include "node.hpp"

// -------------------
Const::Const(double v)
{
	val= v;
}

int Const::type(void) const
{
	return NUM;
}

double Const::eval( double x ) const
{
	return val;
}

string Const::write(void) const
{
	ostringstream out;
	out << val;
	return out.str();
}

Node* Const::deriv(void) const
{
	return new Const(0.0);
}

Node* Const::copy(void) const
{
	return new Const(val);
}

// -------------------

Var::Var(void)
{
}

int Var::type(void) const
{
	return VAR;
}

double Var::eval( double x ) const
{
	return x;
}

string Var::write(void) const
{
	return "x";
}

Node* Var::deriv(void) const
{
	return new Const(1.0);
}

Node* Var::copy(void) const
{
	return new Var();
}

// -------------------

Func::Func(Node *r)
{
	rgt= r;
}

// -------------------

Chs::Chs(Node *r) : Func(r)
{
}

int Chs::type(void) const
{
	return CHS;
}

double Chs::eval( double x ) const
{
	return - rgt->eval(x);
}

string Chs::write(void) const
{
	string r= rgt->write();
	int ty= rgt->type();
	if (ty==ADD || ty==SUB || ty==CHS)
		return "- (" + r + ")";
	else
		return "-" + r;
}

Node* Chs::deriv(void) const
{
	return new Chs(rgt->deriv());
}

Node* Chs::copy(void) const
{
	return new Chs( rgt->copy() );
}

// -------------------

OpBin::OpBin(Node *l, Node *r)
{
	lft= l;
	rgt= r;
}

// -------------------

Add::Add(Node* l, Node* r) : OpBin(l,r)
{
}

int Add::type(void) const
{
	return ADD;
}

double Add::eval( double x ) const
{
	return  lft->eval(x) + rgt->eval(x);
}

string Add::write(void) const
{
	string r= rgt->write();
	string str= lft->write() + "+";
	if (rgt->type()==CHS)
		str= str + "+ (" + r + ")";
	else
		str= str + r;
	return str;
}

Node* Add::deriv(void) const
{
	return new Add( lft->deriv(), rgt->deriv() );
}

Node* Add::copy(void) const
{
	return new Add(lft->copy(), rgt->copy());
}


// -------------------

Sub::Sub(Node* l, Node* r) : OpBin(l,r)
{
}

int Sub::type(void) const
{
	return SUB;
}

double Sub::eval( double x ) const
{
	return  lft->eval(x) - rgt->eval(x);
}

string Sub::write(void) const
{
	string r=  rgt->write();
	if (rgt->type()==ADD || rgt->type()==SUB || rgt->type()==CHS)
		r= "(" + r + ")";
	return lft->write() + "-" + r;
}

Node* Sub::deriv(void) const
{
	return new Sub( lft->deriv(), rgt->deriv() );
}

Node* Sub::copy(void) const
{
	return new Sub(lft->copy(), rgt->copy());
}


// -------------------

Mul::Mul(Node* l, Node* r) : OpBin(l,r)
{
}

int Mul::type(void) const
{
	return MUL;
}

double Mul::eval( double x ) const
{
	return  lft->eval(x) * rgt->eval(x);
}

string Mul::write(void) const
{
	int lt= lft->type();
	string l=  lft->write();
	if (lt==ADD || lt==SUB || lt==CHS)
		l= "(" + l + ")";

	int rt= rgt->type();
	string r=  rgt->write();
	if (rt==ADD || rt==SUB || rt==CHS)
		r= "(" + r + ")";

	return l + "*" + r;
}

Node* Mul::deriv(void) const
{
	return 
		new Add(
			new Mul( lft->copy(), rgt->deriv() ),
			new Mul( lft->deriv(), rgt->copy() )
			);
}

Node* Mul::copy(void) const
{
	return new Mul(lft->copy(), rgt->copy());
}

// -------------------

Div::Div(Node* l, Node* r) : OpBin(l,r)
{
}

int Div::type(void) const
{
	return DIV;
}

double Div::eval( double x ) const
{
	return  lft->eval(x) / rgt->eval(x);
}

string Div::write(void) const
{
	int lt= lft->type();
	string l=  lft->write();
	if (lt==ADD || lt==SUB || lt==CHS)
		l= "(" + l + ")";

	int rt= rgt->type();
	string r=  rgt->write();
	if (rt==ADD || rt==SUB || rt==CHS || rt==MUL || rt== DIV)
		r= "(" + r + ")";

	return l + "/" + r;
}

Node* Div::deriv(void) const
{
	return 
		new Div(	
			new Sub(
				new Mul( rgt->copy(), lft->deriv() ),
				new Mul( lft->copy(), rgt->deriv() )
				),
			new Pow(
					rgt->copy(),
					new Const(2)
				)
			);
}

Node* Div::copy(void) const
{
	return new Div(lft->copy(), rgt->copy());
}

// -------------------

Pow::Pow(Node* l, Node* r) : OpBin(l,r)
{
}

int Pow::type(void) const
{
	return POW;
}

double Pow::eval( double x ) const
{
	return  pow(lft->eval(x), rgt->eval(x));
}

string Pow::write(void) const
{
	int lt= lft->type();
	string l=  lft->write();
	if (lt==ADD || lt==SUB || lt==CHS || lt==MUL || lt==DIV || lt==POW )
		l= "(" + l + ")";

	int rt= rgt->type();
	string r=  rgt->write();
	if (rt==ADD || rt==SUB || rt==CHS || rt==MUL || rt== DIV)
		r= "(" + r + ")";

	return l + "^" + r;
}

Node* Pow::deriv(void) const
{
	Node *base= lft;
	Node *expo= rgt;
	return 
		new Add(
			new Mul(
				expo->copy(),
				new Mul(
						new Pow(
							base->copy(),
							new Sub(
								expo->copy(),
								new Const(1.0)
							)
						),
						base->deriv()
					)
			),
			new Mul (
				new Mul( 
					new Pow( base->copy(), expo->copy() ),
					new Log( base->copy() )
				),
				expo->deriv()
			)
		);
}

Node* Pow::copy(void) const
{
	return new Pow(lft->copy(), rgt->copy());
}

// -------------------

Exp::Exp(Node* r) : Func(r)
{
}

int Exp::type(void) const
{
	return EXP;
}

double Exp::eval( double x ) const
{
	return  exp(rgt->eval(x));
}

string Exp::write(void) const
{
	return "exp(" + rgt->write() + ")"; 
}

Node* Exp::deriv(void) const
{
	return new Mul(
				new Exp( rgt->copy() ),
				rgt->deriv()
				);
}

Node* Exp::copy(void) const
{
	return new Exp(rgt->copy());
}

// -------------------

Log::Log(Node* r) : Func(r)
{
}

int Log::type(void) const
{
	return LOG;
}

double Log::eval( double x ) const
{
	return  log(rgt->eval(x));
}

string Log::write(void) const
{
	return "log(" + rgt->write() + ")"; 
}

Node* Log::deriv(void) const
{
	return new Div(
				rgt->deriv(),
				new Log( rgt->copy() )
				);
}

Node* Log::copy(void) const
{
	return new Log(rgt->copy());
}

